#ifndef SHARK_ALGORITHMS_DIRECT_SEARCH_CMATest_H
#define SHARK_ALGORITHMS_DIRECT_SEARCH_CMATest_H

#include <shark/Algorithms/AbstractSingleObjectiveOptimizer.h>
#include <shark/Statistics/Distributions/MultiVariateNormalDistribution.h>
#include <shark/Algorithms/DirectSearch/Individual.h>
#include <shark/Algorithms/DirectSearch/FitnessExtractor.h>
#include <shark/Algorithms/DirectSearch/Operators/Selection/ElitistSelection.h>
namespace shark {
	class CMATest : public AbstractSingleObjectiveOptimizer<RealVector >
	{
	public:
		double chi( unsigned int n ) {
			return( std::sqrt( static_cast<double>( n ) )*(1. - 1./(4.*n) + 1./(21.*n*n)) );
		}
		/**
		* \brief Default c'tor.
		*/
		CMATest(){}

		/// \brief From INameable: return the class name.
		std::string name() const
		{ return "CMA-ES"; }

		/**
		* \brief Calculates lambda for the supplied dimensionality n.
		*/
		static unsigned suggestLambda( unsigned int dimension ) {
			unsigned lambda = unsigned( 4. + ::floor( 3. * ::log( static_cast<double>( dimension ) ) ) ); // eq. (44)
			// heuristic for small search spaces
			lambda = std::max<unsigned int>( 5, std::min( lambda, dimension ) );
			return lambda ;
		}
		void read( InArchive & archive ){}
		void write( OutArchive & archive ) const{}

		using AbstractSingleObjectiveOptimizer<RealVector >::init;
		void init( ObjectiveFunctionType const& function, SearchPointType const& p){
			unsigned int lambda = CMATest::suggestLambda( p.size() );
			init( function,
				p,
				lambda,
				1.0/std::sqrt(double(p.size()))
			);
		}

		/**
		* \brief Initializes the algorithm for the supplied objective function.
		*/
		void init( 
			ObjectiveFunctionType const& function, 
			SearchPointType const& initialSearchPoint,
			unsigned int lambda, 
			double initialSigma
		) {

			m_numberOfVariables = function.numberOfVariables();
			m_lambda = lambda;
			m_mu = lambda/2;
			m_sigma = initialSigma;

			m_mean.resize( m_numberOfVariables );
			m_evolutionPathC.resize( m_numberOfVariables );
			m_evolutionPathSigma.resize( m_numberOfVariables );
			m_mutationDistribution.resize( m_numberOfVariables );
			m_mean.clear();
			m_evolutionPathC.clear();
			m_evolutionPathSigma.clear();
			
			m_feasible = lambda;
				
			//weighting of the k-best individuals
			m_weights.resize(m_mu);
			for (unsigned int i = 0; i < m_mu; i++)
				m_weights(i) = ::log(m_mu + 0.5) - ::log(1. + i); // eq. (45)
			m_weights /= sum(m_weights); // eq. (45)
			m_muEff = 1. / sum(sqr(m_weights)); // equal to sum(m_weights)^2 / sum(sqr(m_weights))

			// Step size control
			m_cSigma = (m_muEff + 2.)/(m_numberOfVariables + m_muEff + 5.); // eq. (46)
			m_dSigma = 1. + 2. * std::max(0., ::sqrt((m_muEff-1.)/(m_numberOfVariables+1)) - 1.) + m_cSigma; // eq. (46)

			m_cC = (4. + m_muEff / m_numberOfVariables) / (m_numberOfVariables + 4. +  2 * m_muEff / m_numberOfVariables); // eq. (47)
			m_c1 = 2 / (sqr(m_numberOfVariables + 1.3) + m_muEff); // eq. (48)
			double alphaMu = 2.;
			m_cMu = std::min(1. - m_c1, alphaMu * (m_muEff - 2. + 1./m_muEff) / (sqr(m_numberOfVariables + 2) + alphaMu * m_muEff / 2)); // eq. (49)
			m_mean = initialSearchPoint;
			m_best.point = initialSearchPoint; // CI: you can argue about this, as the point is not evaluated
			m_best.value = function(initialSearchPoint); //OK: evaluating performance of first point :P
		}

		/**
		* \brief Executes one iteration of the algorithm.
		*/
		void step(ObjectiveFunctionType const& function){

			std::vector< Individual<RealVector, double, RealVector> > offspring( m_lambda );

			m_feasible = 0;
			for( unsigned int i = 0; i < offspring.size(); i++ ) {
				MultiVariateNormalDistribution::result_type sample = m_mutationDistribution();
				offspring[i].chromosome() = sample.second;
				offspring[i].searchPoint() = m_mean + m_sigma * sample.first;
				
				if( function.isFeasible( offspring[i].searchPoint() ) ) {
					offspring[i].unpenalizedFitness() = function.eval( offspring[i].searchPoint() );
					offspring[i].penalizedFitness() = offspring[i].unpenalizedFitness();
					++m_feasible;
				}else{
					offspring[i].unpenalizedFitness() = 1e100;
					offspring[i].penalizedFitness() = 1e100;
				}
			}

			// Selection
			std::vector< Individual<RealVector, double, RealVector> > parents( m_mu );
			ElitistSelection<FitnessExtractor> selection;
			selection(offspring.begin(),offspring.end(),parents.begin(), parents.end());
			// Strategy parameter update
			updateStrategyParameters( parents,function );
			if(offspring[ 0 ].penalizedFitness() < 1e100){
				m_best.point= offspring[ 0 ].searchPoint();
				m_best.value= offspring[ 0 ].unpenalizedFitness();
			}
		}
		
		double sigma() const {
			return m_sigma;
		}
		
		void setSigma(double sigma) {
			m_sigma = sigma;
		}
		
		unsigned int feasible()const{
			return m_feasible;
		}
		
		
		/**
		 * \brief Returns a eigenvectors of covariance matrix (not considering step size)
		 */
		RealVector const& eigenValues() const {
			return m_mutationDistribution.eigenValues();
		}



	private:
		/**
		* \brief Updates the strategy parameters based on the supplied offspring population.
		*/
		void updateStrategyParameters( const std::vector<Individual<RealVector, double, RealVector> > & offspring,ObjectiveFunctionType const& function ) {
			RealVector z(m_numberOfVariables,0.0);
			RealVector m(m_numberOfVariables,0.0);
			for(std::size_t i = 0; i != m_mu; ++i){
				noalias(z) += m_weights(i)*offspring[i].chromosome();
				noalias(m) += m_weights(i)*offspring[i].searchPoint();
			}
			RealVector y = (m - m_mean) / m_sigma;

			// Covariance matrix update
			RealMatrix& C = m_mutationDistribution.covarianceMatrix();
			RealMatrix Z( m_numberOfVariables, m_numberOfVariables, 0.0); // matric for rank-mu update
			for( unsigned int i = 0; i < m_mu; i++ ) {
				noalias(Z) += m_weights( i ) * blas::outer_prod(
					offspring[i].searchPoint() - m_mean,
					offspring[i].searchPoint() - m_mean
				);
			}

			m_evolutionPathC = (1. - m_cC ) * m_evolutionPathC + std::sqrt( m_cC * (2. - m_cC) * m_muEff ) * y; // eq. (42)
			noalias(C) = (1.-m_c1 - m_cMu) * C + m_c1 * blas::outer_prod( m_evolutionPathC, m_evolutionPathC ) + m_cMu * 1./sqr( m_sigma ) * Z; // eq. (43)

			// Step size update
			RealVector CInvY = blas::prod( m_mutationDistribution.eigenVectors(), z ); // C^(-1/2)y = Bz
			m_evolutionPathSigma = (1. - m_cSigma)*m_evolutionPathSigma + std::sqrt( m_cSigma * (2. - m_cSigma) * m_muEff ) * CInvY; // eq. (40)
			m_sigma *= std::exp( (m_cSigma / m_dSigma) * (norm_2(m_evolutionPathSigma)/ chi( m_numberOfVariables ) - 1.) ); // eq. (39)

			
			// update mutation distribution
			m_mutationDistribution.update();
			
			//mean update
			RealVector newMean=m;
			double stepLength = 1.0;
			while(!function.isFeasible(newMean)){
				stepLength/=1.5;
				newMean = m_mean+stepLength*(m-m_mean);
			}
			m_mean = newMean;
		}
	
		unsigned int m_numberOfVariables; ///< Stores the dimensionality of the search space.
		unsigned int m_mu; ///< The size of the parent population.
		unsigned int m_lambda; ///< The size of the offspring population, needs to be larger than mu.
		unsigned int m_feasible;
		
		double m_sigma;
		double m_cC; 
		double m_c1; 
		double m_cMu; 
		double m_cSigma;
		double m_dSigma;
		double m_muEff;

		RealVector m_mean;
		RealVector m_weights;

		RealVector m_evolutionPathC;
		RealVector m_evolutionPathSigma;

		MultiVariateNormalDistribution m_mutationDistribution;
	};
}

#endif
