/*!
 * 
 *
 * \brief       Example of an experiment using the CMA-ES on several benchmark functions
 *
 * \author      O.Krause
 * \date        2014
 *
 * \par Copyright 1995-2014 Shark Development Team
 * 
 * <BR><HR>
 * This file is part of Shark.
 * <http://image.diku.dk/shark/>
 * 
 * Shark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Shark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Shark.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// Implementations of the xNES
#include "CMA.h"
#include "xCMA.h"
#include <shark/Algorithms/DirectSearch/CMA.h>

//benchmark functions
#include <shark/ObjectiveFunctions/Benchmarks/ConstrainedSphere.h>

//timing
#include <shark/Core/Timer.h>

//misc
#include <shark/Core/utility/functional.h>
#include <boost/lexical_cast.hpp>
#include <fstream>

		
using namespace shark;

typedef boost::shared_ptr<SingleObjectiveFunction > Function;
template<class Optimizer>
void evaluate(std::size_t numTrials,double targetValue,std::size_t numDimensions){
	Optimizer cma;
	double failure = 1e100;
	double maxIter= 30000;
	
	std::size_t lambda = unsigned( 4. + ::floor( 3. * ::log( static_cast<double>( numDimensions ) ) ) );
	lambda = std::max<std::size_t>( 5, std::min( lambda, numDimensions ) );
	
	std::ofstream str(("results/"+cma.name()+"-Constrained"+boost::lexical_cast<std::string>(numDimensions)+".txt").c_str());
	std::cout<< "# optimizer: "<<cma.name()<<std::endl;
	str<< "# optimizer: "<<cma.name()<<std::endl;
	std::cout<<"#m median lowerQ upperQ successTrials "<<std::endl;
	str<<"#m median lowerQ upperQ successTrials "<<std::endl;
	for(std::size_t d = 1; d != 6; ++d){
		std::size_t m = (1<<d);
		if(m > numDimensions/2) break;
		
		std::vector<double> results(numTrials,0.0);
		ConstrainedSphere function(numDimensions,m);
		for(std::size_t trial = 0; trial != numTrials; ++trial){
			std::cout<<"\r"<<trial<<"/"<<numTrials<<std::flush;
			//init function
			Rng::seed(142+trial);
			function.init();
			
			//initialize the optimizer
			cma.init( function );
			//optimize
			while(cma.solution().value > targetValue){
				//check for premature convergence(e.g. local optimum)
				if(max(cma.eigenValues())*cma.sigma()<1.e-18 
				|| max(cma.eigenValues())/min(cma.eigenValues())> 1.e18
				|| results[trial] > maxIter
				){	//failure!
					results[trial]=failure;
					break;
				}
				
				cma.step(function);
				results[trial]+=1;
				
				
			}
		}
		//sort and remove failed trials
		std::sort(results.begin(),results.end());
		while(results.back()==failure) results.pop_back();
		std::cout<<m<<" "
			<<results[0.5*numTrials]<<" "
			<<results[0.25*numTrials]<<" "
			<<results[0.75*numTrials]<<" "
			<<results.size()<<std::endl;
		str<<m<<" "
			<<results[0.5*numTrials]<<" "
			<<results[0.25*numTrials]<<" "
			<<results[0.75*numTrials]<<" "
			<<results.size()<<std::endl;
	}
	
}

int main( int argc, char ** argv ) {

	std::size_t numTrials = 100; // how often the optimization is repeated
	double targetValue = 1.e-12;//target value to reach
	
	//to check against the normal CMA just comment the lines with CMATest and uncomment the lines with CMA
	evaluate<xCMA>(numTrials, targetValue,32);
	evaluate<CMATest>(numTrials, targetValue,32);
	//~ evaluate<CMA>(numTrials, targetValue,32);
	evaluate<xCMA>(numTrials, targetValue,16);
	evaluate<CMATest>(numTrials, targetValue,16);
	//~ evaluate<CMA>(numTrials, targetValue,16);
	
}
