/*!
 * 
 *
 * \brief       Example of an experiment using the CMA-ES on several benchmark functions
 *
 * \author      O.Krause
 * \date        2014
 *
 * \par Copyright 1995-2014 Shark Development Team
 * 
 * <BR><HR>
 * This file is part of Shark.
 * <http://image.diku.dk/shark/>
 * 
 * Shark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Shark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Shark.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// Implementations of the xNES
#include "xCMA.h"
#include <shark/Algorithms/DirectSearch/CMA.h>

//benchmark functions
#include <shark/ObjectiveFunctions/Benchmarks/Rosenbrock.h>
#include <shark/ObjectiveFunctions/Benchmarks/Sphere.h>
#include <shark/ObjectiveFunctions/Benchmarks/Schwefel.h>
#include <shark/ObjectiveFunctions/Benchmarks/Cigar.h>
#include <shark/ObjectiveFunctions/Benchmarks/Discus.h>
#include <shark/ObjectiveFunctions/Benchmarks/Ellipsoid.h>
#include <shark/ObjectiveFunctions/Benchmarks/DiffPowers.h>
#include <shark/ObjectiveFunctions/Benchmarks/RotatedErrorFunction.h>
#include "SharpRidge.h"
#include "ParabRidge.h"
//timing
#include <shark/Core/Timer.h>

//misc
#include <shark/Core/utility/functional.h>
#include <boost/lexical_cast.hpp>
#include <fstream>

		
using namespace shark;

typedef boost::shared_ptr<SingleObjectiveFunction > Function;
template<class Optimizer>
void evaluate(std::vector<Function >const& functions,std::size_t numTrials,RealVector fstop){
	Optimizer cma;
	double failure = 1e100;
	//we print everything on std::cout and in a file
	std::ofstream str(("results/"+cma.name()+".txt").c_str());
	std::cout<< "# optimizer: "<<cma.name()<<std::endl;
	str<< "# optimizer: "<<cma.name()<<std::endl;
	for(std::size_t d = 1; d != 6; ++d){
		std::size_t numDimensions = 2*(1<<d);
		RealVector medianPerformance(functions.size(),0.0);
		RealVector lowerQuantilePerformance(functions.size(),0.0);
		RealVector upperQuantilePerformance(functions.size(),0.0);
		for(std::size_t f = 0; f != functions.size(); ++f){
			std::vector<double> results(numTrials,0.0);
			functions[f]->setNumberOfVariables(numDimensions);
			for(std::size_t trial = 0; trial != numTrials; ++trial){
				//init function
				Rng::seed(142+trial);
				functions[f]->init();
				
				//progress
				std::cout<<"\r"<<functions[f]->name()<<": "<<trial<<std::flush;
				
				//initialize the optimizer
				cma.init( *functions[f] );
				//optimize
				while(cma.solution().value > fstop(f)){
					cma.step(*functions[f]);
					results[trial]+=1;
					//~ std::cout<<results[trial]<<" "<<cma.eigenValues()<<" "<<cma.sigma()<<std::endl;
					//check for premature convergence(e.g. local optimum)
					if(max(cma.eigenValues())*cma.sigma()<1.e-18 
					|| max(cma.eigenValues())/min(cma.eigenValues())> 1.e15
					){
						//failure!
						results[trial]=failure;
						break;
					}
					
				}
				
			}
			//remove failed trials
			std::sort(results.begin(),results.end());
			while(results.back()==failure) results.pop_back();
			std:sort(results.begin(),results.end());
			medianPerformance(f) = results[0.5*numTrials];
			lowerQuantilePerformance(f) = results[0.25*numTrials];
			upperQuantilePerformance(f) = results[0.75*numTrials];
			std::cout<<"\r                                             \r";
		}
		
		std::size_t lambda = unsigned( 4. + ::floor( 3. * ::log( static_cast<double>( numDimensions ) ) ) );
		lambda = std::max<std::size_t>( 5, std::min( lambda, numDimensions ) );
		
		std::cout<<"\r                              \r";
		std::cout<<numDimensions<<" ";
		str<<numDimensions<<" ";
		for(std::size_t f = 0; f != functions.size(); ++f){
			std::cout.precision( 7 );
			str.precision( 7 );
			std::cout<<lambda*medianPerformance(f)<<" ";
			str<<lambda*medianPerformance(f)<<" "
			<<lambda*lowerQuantilePerformance(f)<<" "
			<<lambda*upperQuantilePerformance(f)<<" ";
		}
		std::cout<<std::endl;
		str<<std::endl;
	}
	std::cout<<"\n";
	std::cout<<"\r                              \r";
	
}

int main( int argc, char ** argv ) {

	std::size_t numTrials = 100; // how often the optimization is repeated
	double targetValue = 1.e-14;//target value to reach
	
	//assortment of test functions
	std::vector<Function > functions;
	functions.push_back(Function(new SharpRidge()));
	functions.push_back(Function(new ParabRidge()));
	functions.push_back(Function(new Rosenbrock()));
	functions.push_back(Function(new Sphere()));
	functions.push_back(Function(new Schwefel()));
	functions.push_back(Function(new Discus(4,1.e-6)));
	functions.push_back(Function(new Cigar(4,1.e-6)));
	functions.push_back(Function(new Ellipsoid(4,1.e-6)));
	functions.push_back(Function(new DiffPowers()));
	
	
	RealVector fstop(functions.size(),targetValue);
	fstop(0) = -1000;
	fstop(1) = -1000;
	//add rootated versions
	//~ std::size_t baseFunctions = functions.size();
	//~ for(std::size_t i = 0; i != baseFunctions; ++i){
		//~ functions.push_back(Function(new RotatedObjectiveFunction(functions[i].get())));
	//~ }
	std::cout<<"#function evaluations\n";
	std::cout<<"#d ";
	for(std::size_t f = 0; f != functions.size(); ++f){
		std::cout<<functions[f]->name() <<" ";
	}
	std::cout<<std::endl;
	evaluate<xCMA>(functions,numTrials,fstop);
	evaluate<CMA>(functions,numTrials, fstop);
}
