/*!
 * 
 *
 * \brief       Example of an experiment using the CMA-ES on several benchmark functions
 *
 * \author      O.Krause
 * \date        2014
 *
 * \par Copyright 1995-2014 Shark Development Team
 * 
 * <BR><HR>
 * This file is part of Shark.
 * <http://image.diku.dk/shark/>
 * 
 * Shark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Shark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Shark.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// Implementations of the xNES
#include "xCMA.h"
#include "CMA.h"
#include <shark/Algorithms/DirectSearch/CMA.h>

//benchmark functions
#include <shark/ObjectiveFunctions/Benchmarks/Rosenbrock.h>
#include <shark/ObjectiveFunctions/Benchmarks/Sphere.h>
#include <shark/ObjectiveFunctions/Benchmarks/Schwefel.h>
#include <shark/ObjectiveFunctions/Benchmarks/Cigar.h>
#include <shark/ObjectiveFunctions/Benchmarks/Discus.h>
#include <shark/ObjectiveFunctions/Benchmarks/Ellipsoid.h>
#include <shark/ObjectiveFunctions/Benchmarks/DiffPowers.h>
#include <shark/ObjectiveFunctions/Benchmarks/RotatedErrorFunction.h>
#include <shark/ObjectiveFunctions/Benchmarks/ConstrainedSphere.h>

//timing
#include <shark/Core/Timer.h>

//misc
#include <shark/Core/utility/functional.h>
#include <boost/lexical_cast.hpp>
#include <fstream>

		
using namespace shark;

typedef boost::shared_ptr<SingleObjectiveFunction > Function;
template<class Optimizer>
void evaluate(std::vector<Function >const& functions,double targetValue){
	Optimizer cma;
	std::cout<< "# optimizer: "<<cma.name()<<std::endl;
	for(std::size_t f = 0; f != functions.size(); ++f){
		//init function
		Rng::seed(142+f);
		functions[f]->init();
		
		unsigned int lambda = Optimizer::suggestLambda(functions[f]->numberOfVariables());
		
		//initialize the optimizer
		cma.init( *functions[f] );
		
		std::string fileName="results/"+cma.name()+"-"+functions[f]->name()+".txt";
		std::ofstream str(fileName.c_str());
		str<<"#iteration value sigma eigenvalues \n";  
		//optimize
		unsigned int iter = 0;
		while(cma.solution().value > targetValue){
			str.precision( 7 );
			str<<lambda*iter<<" "<<cma.solution().value<<" "<<cma.sigma()<<" ";
			RealVector eig=cma.eigenValues();
			for(std::size_t i = 0; i != eig.size(); ++i){
				str<<eig(i)<<" ";
			}
			str<<"\n";
			cma.step(*functions[f]);
			++iter;
			if(iter == 20000) break;
		}
		str<<lambda*iter<<" "<<cma.solution().value<<" "<<cma.sigma()<<" ";
		RealVector eig=cma.eigenValues();
		for(std::size_t i = 0; i != eig.size(); ++i){
			str<<eig(i)<<" ";
		}
		str<<"\n";
	}
}

int main( int argc, char ** argv ) {

	double targetValue = 1.e-13;
	std::size_t numDimensions = 32;
	
	//not included in the paper: a few plots of various functions
	//~ //assortment of test functions
	//~ std::vector<Function > functions;
	//~ functions.push_back(Function(new Rosenbrock(numDimensions)));
	//~ functions.push_back(Function(new Sphere(numDimensions)));
	//~ functions.push_back(Function(new Schwefel(numDimensions)));
	//~ functions.push_back(Function(new Discus(numDimensions,1.e-6)));
	//~ functions.push_back(Function(new Cigar(numDimensions,1.e-6)));
	//~ functions.push_back(Function(new Ellipsoid(numDimensions,1.e-6)));
	//~ functions.push_back(Function(new DiffPowers(numDimensions)));
	
	//~ evaluate<xCMA>(functions,targetValue);
	//~ evaluate<CMA>(functions,targetValue);
	
	
	std::vector<Function > constrainedFunctions;
	constrainedFunctions.push_back(Function(new ConstrainedSphere(numDimensions,4)));
	
	evaluate<CMATest>(constrainedFunctions,targetValue);// adaptedimplementation for constraint handling
	evaluate<xCMA>(constrainedFunctions,targetValue);
	
}
