#ifndef SHARK_ALGORITHMS_DIRECT_SEARCH_xCMA_H
#define SHARK_ALGORITHMS_DIRECT_SEARCH_xCMA_H

#include <shark/Algorithms/AbstractSingleObjectiveOptimizer.h>
#include <shark/Statistics/Distributions/MultiVariateNormalDistribution.h>
#include <shark/Algorithms/DirectSearch/Individual.h>


namespace shark {
	RealMatrix expm(RealMatrix M){
		RealMatrix Q;
		RealVector D;
		diag(M)+=1000;
		eigensymm( M, Q, D );//todo: negative eigenvalues not allowed
		D-=1000;
		noalias(D)=exp(D/2);
		//~ std::cout<<sqr(D)<<std::endl;
		for(std::size_t i = 0; i != M.size1(); ++i){
			column(Q,i) *= D(i);
		}
		return prod(Q,trans(Q));
	}
	
	double chi( unsigned int n ) {
		return( std::sqrt( static_cast<double>( n ) )*(1. - 1./(4.*n) + 1./(21.*n*n)) );
	}
	
	class xCMA : public AbstractSingleObjectiveOptimizer<RealVector >
	{
	public:
		/**
		* \brief Default c'tor.
		*/
		xCMA(){}

		/// \brief From INameable: return the class name.
		std::string name() const
		{ return "xCMA-ES"; }

		/**
		* \brief Calculates lambda for the supplied dimensionality n.
		*/
		static unsigned suggestLambda( unsigned int dimension ) {
			unsigned lambda = unsigned( 4. + ::floor( 3. * ::log( static_cast<double>( dimension ) ) ) ); // eq. (44)
			// heuristic for small search spaces
			lambda = std::max<unsigned int>( 5, std::min( lambda, dimension ) );
			return lambda ;
		}
		void read( InArchive & archive ){}
		void write( OutArchive & archive ) const{}

		using AbstractSingleObjectiveOptimizer<RealVector >::init;
		void init( ObjectiveFunctionType const& function, SearchPointType const& p){
			unsigned int lambda = xCMA::suggestLambda( p.size() );
			init( function,
				p,
				lambda,
				1.0/std::sqrt(double(p.size()))
			);
		}

		/**
		* \brief Initializes the algorithm for the supplied objective function.
		*/
		void init( 
			ObjectiveFunctionType const& function, 
			SearchPointType const& initialSearchPoint,
			unsigned int lambda, 
			double initialSigma
		){
			m_numberOfVariables = function.numberOfVariables();
			m_lambda = lambda;
			m_sigma = initialSigma;

			m_mean.resize( m_numberOfVariables );
			m_evolutionPathC.resize( m_numberOfVariables );
			m_evolutionPathSigma.resize( m_numberOfVariables );
			m_B = blas::identity_matrix<double>( m_numberOfVariables );
			m_eig = blas::repeat(1.0,m_numberOfVariables);
			m_mean.clear();
			m_evolutionPathC.clear();
			m_evolutionPathSigma.clear();
			m_weights.resize(m_lambda);
			m_feasible = lambda;
			
			m_mean = initialSearchPoint;
			m_best.point = initialSearchPoint;
			m_best.value = function(initialSearchPoint);
		}

		/**
		* \brief Executes one iteration of the algorithm.
		*/
		void step(ObjectiveFunctionType const& function){
			std::vector< Individual<RealVector, double, RealVector> > offspring( m_lambda );

			m_feasible = 0;
			for( unsigned int i = 0; i < offspring.size(); i++ ) {
				sample(offspring[i].chromosome(),offspring[i].searchPoint());
				if( function.isFeasible( offspring[i].searchPoint() ) ) {
					offspring[i].unpenalizedFitness() = function.eval( offspring[i].searchPoint() );
					offspring[i].penalizedFitness() = offspring[i].unpenalizedFitness();
					++m_feasible;
				}else{
					offspring[i].unpenalizedFitness() = 1e100;
					offspring[i].penalizedFitness() = 1e100;
				}
			}
			
			//weighting of the k-best individuals
			m_weights.clear();
			for (unsigned int i = 0; i < m_lambda; i++)
				m_weights(i) = std::max(0.0,std::log(m_lambda/2 + 1.0) - std::log(1. + i));
			
			double sumOfPos= sum(m_weights);
			for (unsigned int i = m_feasible; i < m_lambda; i++){
				m_weights(i) -= 0.4/m_lambda*sumOfPos;
			}
			m_weights /= sum(abs(m_weights));
			m_muEff = 1. / sum(sqr(m_weights)); // equal to sum(m_weights)^2 / sum(sqr(m_weights))
			//~ std::cout<<feasible<<" "<<m_muEff<<std::endl;
			m_weights -= sum(m_weights)/m_lambda;//make mean-free
			//~ std::cout<<m_weights<<std::endl;
			
			// Step size control
			double d=double(m_numberOfVariables);
			m_cSigma = (m_muEff + 2.)/(d + m_muEff + 5.); // eq. (46)
			m_dSigma = 1. + 2. * std::max(0., ::sqrt((m_muEff-1.)/(d+1)) - 1.) + m_cSigma; // eq. (46)

			m_cC = (4. + m_muEff / d) / (d + 4. +  2 * m_muEff / d); // eq. (47)
			m_c1 = 2 / (sqr(d + 1.3) + m_muEff); // eq. (48)
			double alphaMu = 2.;
			m_cMu = std::min(1. - m_c1, alphaMu * (m_muEff - 2. + 1./m_muEff) / (sqr(d + 2) + alphaMu * m_muEff / 2)); // eq. (49)
			std::sort(offspring.begin(),offspring.end(), Individual<RealVector, double, RealVector>::FitnessOrdering());
			// Strategy parameter update
			updateStrategyParameters( offspring,function);
			
			if(feasible() > 0){
				m_best.point= offspring[ 0 ].searchPoint();
				m_best.value= offspring[ 0 ].unpenalizedFitness();
			}
		}
		
		double sigma() const {
			return m_sigma;
		}
		
		void setSigma(double sigma) {
			m_sigma = sigma;
		}
		
		unsigned int feasible()const{
			return m_feasible;
		}
		
		
		/**
		 * \brief Returns a eigenvectors of covariance matrix (not considering step size)
		 */
		RealVector eigenValues() const {
			return sqr(m_eig);
		}

	private:
		void sample(RealVector &z, RealVector& x){
			z.resize(m_numberOfVariables);
			x.resize(m_numberOfVariables);
			for(std::size_t i = 0; i != m_numberOfVariables; ++i){
				z(i) = Rng::gauss(0,1);
				x(i) = z(i)*m_eig(i)*m_sigma;
			}
			x = m_mean + prod(m_B,x);
		}
		/**
		* \brief Updates the strategy parameters based on the supplied offspring population.
		*/
		void updateStrategyParameters( const std::vector<Individual<RealVector, double, RealVector> > & offspring,ObjectiveFunctionType const& function ){
			RealVector zBar(m_numberOfVariables,0.0);
			RealVector m(m_numberOfVariables,0.0);
			double sumWeights = sum(m_weights);
			for(std::size_t i = 0; i != m_lambda; ++i){
				noalias(zBar) += (m_weights(i)+(1.0-sumWeights)/m_lambda)*offspring[i].chromosome();
				noalias(m) += (m_weights(i)+(1.0-sumWeights)/m_lambda)*offspring[i].searchPoint();
			}
			

			//let C=BD^2B^T
			//we sample x=mu+C^1/2y = mu + BDB^Ty=mu + BDz, so z=B^Ty
			// Step size update
			RealVector CInvYBar = blas::prod( m_B, zBar ); // y_bar = Bz_bar
			m_evolutionPathSigma = (1. - m_cSigma)*m_evolutionPathSigma + std::sqrt( m_cSigma * (2. - m_cSigma) * m_muEff ) * CInvYBar; // etq. (40)

			//evolution path
			noalias(m_evolutionPathC) = (1. - m_cC ) * m_evolutionPathC + std::sqrt( m_cC * (2. - m_cC) * m_muEff )/m_sigma *  (m - m_mean);
			RealVector p = prod(trans(m_B),m_evolutionPathC);
			p /= m_eig;
			
			// Covariance matrix update
			//Z=sum_i w_i z_i z_i^T=B^T sum_i w_i y_i y_i^T B
			RealMatrix Z( m_numberOfVariables, m_numberOfVariables, 0.0); // matric for rank-mu update
			for( unsigned int i = 0; i < m_lambda; i++ ) {
				noalias(Z) += m_cMu * m_weights( i ) * blas::outer_prod(
					offspring[i].chromosome(),
					offspring[i].chromosome()
				);
			}

			// cumulative step size adaptation
			m_sigma *= std::exp( (m_cSigma / m_dSigma) * (norm_2(m_evolutionPathSigma)/ chi( m_numberOfVariables ) - 1.) ); // eq. (39)

			noalias(Z) += m_c1*outer_prod(p,p);
			diag(Z)-= m_c1+sumWeights;
			
			
			//compute updated covariance matrix
			//as C'=Aexp(Z)A^T=(Aexp(Z/2))(Aexp(Z/2))^T with C=AA^T, A=BD
			//this gives Aexp(Z)A^T= BDexp(Z) DB^T = BDexp(B^T  sum_i w_i y_i y_i^TB)DB^T
			// = BDB^T exp( sum_i w_i y_i y_i^T)BDB^T
			// = C^1/2 exp( sum_i w_i y_i y_i^T)C^1/2
			//which is the target update!
			RealMatrix A= m_B;
			for(std::size_t i = 0; i != m_numberOfVariables; ++i){
				column(A,i) *= m_eig(i);
			}
			Z/=2.0;
			
			
			//normalize determinant to 1
			//~ double trZ =trace(Z);
			//~ Z = expm(Z)*exp(-trZ/m_numberOfVariables);
			Z = expm(Z);
			A = prod(A,Z);
			RealMatrix Cprime= prod(A,trans(A));
			// update mutation distribution
			eigensymm( Cprime, m_B, m_eig );
			noalias(m_eig) = sqrt(m_eig);
			
			RealVector newMean=m;
			double stepLength = 1.0;
			while(!function.isFeasible(newMean)){
				stepLength/=1.5;
				newMean = m_mean+stepLength*(m-m_mean);
			}
			m_mean = newMean;
			
		}
	
		unsigned int m_numberOfVariables; ///< Stores the dimensionality of the search space.
		unsigned int m_lambda; ///< The size of the offspring population, needs to be larger than mu.
		unsigned int m_feasible;

		double m_sigma;
		double m_cC; 
		double m_c1; 
		double m_cMu; 
		double m_cSigma;
		double m_dSigma;
		double m_muEff;

		RealVector m_mean;
		RealVector m_weights;

		RealVector m_evolutionPathC;
		RealVector m_evolutionPathSigma;

		RealVector m_eig;
		RealMatrix m_B;
	};
}

#endif
